/*
  Exception handling
  Copyright (C) 1999  Hardy Kahl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ------------------------------------------------------------------------------
//  Description      : Exception handling
// ------------------------------------------------------------------------------
//  Remarks          :
//    The additional info _should_ be language independent.
//    These exceptions _must_ be caught by reference.
//
// ------------------------------------------------------------------------------
//  (c) Copyright 1996 by iXiONmedia, all rights reserved.
// ------------------------------------------------------------------------------




#ifndef EXCEPT_H
#define EXCEPT_H




#include "base.h"




typedef unsigned int TErrorCode;




// Throw macro ----------------------------------------------------------------
#define EX_THROW(TYPE,CODE) \
  throw x##TYPE##Exception(CODE,NULL,__FILE__,__LINE__);
#define EX_THROWINFO(TYPE,CODE,INFO) \
  throw x##TYPE##Exception(CODE,(char const *) INFO,__FILE__,__LINE__);
#define EX_CATCHCODE(TYPE,CODE,HANDLER) \
  catch (x##TYPE##Exception &ex) { \
    if (ex.Error != CODE) throw; HANDLER \
  }
#define EX_CONVERT(TYPE,CODE,DESTTYPE,DESTCODE) \
  catch (x##TYPE##Exception &ex) { \
    if (ex.Error != CODE) throw; \
    throw x##DESTTYPE##Exception(DESTCODE,ex.Info,__FILE__,__LINE__); \
  }




// xBaseException -------------------------------------------------------------
struct xBaseException {
  TErrorCode            Error;
  char                  *Module;
  TIndexD               Line;
  char                  *Category;
  TBool			HasInfo;
  static char           Info[];

  xBaseException(TErrorCode error,char const *info = 0,char *module = 0,
    TIndexD line = 0,char *category = 0);
  char *GetDescription(char *const stringbuffer) const;
  virtual char *GetText() const = 0;
};




#endif
