/*
  RAM program definitions
  Copyright (C) 1999  Hardy Kahl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ----------------------------------------------------------------------------
//  Description      : RAM program definitions.
// ----------------------------------------------------------------------------




#include "prgram.h"
#include "xstrio.h"




// Plain text rendering table -------------------------------------------------
#define T_INVALID_LINE          "Invalid line number"




static char *(PlainText[]) =
  { T_INVALID_LINE };




// xRAMException --------------------------------------------------------------
xRAMProgException::xRAMProgException(TErrorCode error, TIndex line, xString &info)
: xBaseException(error, NULL, NULL, 0, "PRR")
{
  HasInfo = TRUE;
  try {
    xString temp = "In line ";
    temp += xString::Unsigned2Dec(line);
    temp += " : ";
    temp += info;
    strcpy(Info, temp);
  }
  catch (...) { }
}




char *xRAMProgException::GetText() const
{
  return PlainText[Error];
}




// xRAMProgram : --------------------------------------------------------------
xRAMProgram::xRAMProgram(void)
: xProgram<xString>()
{
}




void xRAMProgram::Read(istream &stream)
{
  xString line;
  xString number;
  TIndex  i = 1;
  TIndex  val;

  Empty();
  while (stream >> line)
  {
    number.Clear();
    while (~line && (line[0] != ' '))
    {
      number += line[0];
      line.DeleteThis(0, 1);
    }
    while (~line && (line[0] == ' ')) line.DeleteThis(0, 1);

    try
    {
      val = number.EvalUnsigned();
    }
    catch (...)
    {
      xString temp;
      try
      {
        temp = "expected '";
        temp += xString::Unsigned2Dec(i);
        temp += "' but found '";
        temp += number;
        temp += "'";
      }
      catch (...) { }
      throw xRAMProgException(ECRAMP_INVALID_LINE, i, temp);
    }

    if (val != i)
    {
      xString temp;
      try
      {
        temp = "expected '";
        temp += xString::Unsigned2Dec(i);
        temp += "' but found '";
        temp += number;
        temp += "'";
      }
      catch (...) { }
      throw xRAMProgException(ECRAMP_INVALID_LINE, i, temp);
    }

    i++;
    operator+=(line);
  }

  if (~line)
  {
    number.Clear();
    while (~line && (line[0] != ' '))
    {
      number += line[0];
      line.DeleteThis(0, 1);
    }
    while (~line && (line[0] == ' ')) line.DeleteThis(0, 1);

    try
    {
      val = number.EvalUnsigned();
    }
    catch (...)
    {
      xString temp;
      try
      {
        temp = "expected '";
        temp += xString::Unsigned2Dec(i);
        temp += "' but found '";
        temp += number;
        temp += "'";
      }
      catch (...) { }
      throw xRAMProgException(ECRAMP_INVALID_LINE, i, temp);
    }

    if (val != i)
    {
      xString temp;
      try
      {
        temp = "expected '";
        temp += xString::Unsigned2Dec(i);
        temp += "' but found '";
        temp += number;
        temp += "'";
      }
      catch (...) { }
      throw xRAMProgException(ECRAMP_INVALID_LINE, i, temp);
    }

    operator+=(line);
  }
}




void xRAMProgram::Write(ostream &stream)
{
  for (TIndex i=0; i<Items; i++)
  {
    stream << xString::Unsigned2Dec(i+1) << " " << List[i] << endl;
  }
}
