/*
  RAM program definitions
  Copyright (C) 1999  Hardy Kahl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ----------------------------------------------------------------------------
//  Description      : RAM program definitions.
// ----------------------------------------------------------------------------
//  Remarks          :
//    Hardy     18.11.1998      nobody should change this file, except the
//                              whole team decides.
// ----------------------------------------------------------------------------
//  Modifications    :
//    Hardy     18.11.1998      created this file.
// ----------------------------------------------------------------------------




#ifndef PRGRAM_H
#define PRGRAM_H




#include "prgbase.h"
#include "xstring.h"
#include "exbase.h"




// possible errors while execution : ------------------------------------------
#define ECRAMP_INVALID_LINE     0




// xRAMProgException ----------------------------------------------------------
struct xRAMProgException : public xBaseException {
  xRAMProgException(TErrorCode error, TIndex line, xString &info);
  virtual char *GetText() const;
};




// xRAMCode : -----------------------------------------------------------------
class xRAMCode
{
  public:
  static xString Mov_R_R(xString const &param1, xString const &param2)
    { return xString("R") + param1 + " <- R" + param2; }
  static xString Mov_R_RR(xString const &param1, xString const &param2)
    { return xString("R") + param1 + " <- RR" + param2; }
  static xString Mov_RR_R(xString const &param1, xString const &param2)
    { return xString("RR") + param1 + " <- R" + param2; }
  static xString Mov_R_K(xString const &param1, xString const &param2)
    { return xString("R") + param1 + " <- " + param2; }
  static xString Add(xString const &param1, xString const &param2, xString const &param3)
    { return xString("R") + param1 + " <- R" + param2 + " + R" + param3; }
  static xString Sub(xString const &param1, xString const &param2, xString const &param3)
    { return xString("R") + param1 + " <- R" + param2 + " - R" + param3; }
  static xString Goto(xString const &param1)
    { return xString("GOTO ") + param1; }
  static xString If_Null(xString const &param1, xString const &param2)
    { return xString("IF R") + param1 + " = 0 GOTO " + param2; }
  static xString If_Not_Null(xString const &param1, xString const &param2)
    { return xString("IF R") + param1 + " # 0 GOTO " + param2; }
  static xString Stop()
    { return xString("STOP"); }
};




// xRAMPrograms : -------------------------------------------------------------
class xRAMProgram : public xProgram<xString>
{
  public:
  xRAMProgram(void);

  virtual void Read(istream &stream);
  virtual void Write(ostream &stream);
};




inline istream &operator>>(istream &istr, xRAMProgram &prog)
{
  prog.Read(istr);
  return istr;
}




inline ostream &operator<<(ostream &ostr, xRAMProgram &prog)
{
  prog.Write(ostr);
  return ostr;
}




#endif
