/*
  Random Access Machine Simulator
  Copyright (C) 1999  Hardy Kahl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ----------------------------------------------------------------------------
//  Description      : Random Access Machine Simulator.
// ----------------------------------------------------------------------------
//  Remarks          :
//    Hardy     12.11.1998      nobody should change this file, except the
//                              whole team decides.
// ----------------------------------------------------------------------------
//  Modifications    :
//    Hardy     12.11.1998      created this file.
//    Hardy     25.03.1999      fixed bug in void xRAM::Execute(TCount numcmd)
//                                (wrong parameter in exception throw).
//                              added "STOP".
//                              fixed line problem in "xRAMException".
//                              made "BCProgram" and "Register" public.
// ----------------------------------------------------------------------------




#ifndef RAM_H
#define RAM_H




#include "base.h"
#include "prgbc.h"
#include "prgram.h"
#include "register.h"
#include "exbase.h"
#include "xstring.h"




// possible errors while execution : ------------------------------------------
#define ECRAM_INVALID_BCODE     0
#define ECRAM_INVALID_GOTO      1




// xRAMException --------------------------------------------------------------
struct xRAMException : public xBaseException {
  xRAMException(TErrorCode error, TIndex line, xString &info);
  virtual char *GetText() const;
};




// RAM : ----------------------------------------------------------------------
class xRAM
{
  protected:
  TIndex                CommandRegister;
  TBool                 Terminated;

  public:
  xBCProgram            BCProgram;
  xRegisterManager      RegisterManager;

  xRAM(void);

  void Load(xRAMProgram &program);
  void Reset(void);
  TBool HasTerminated(void)
    { return Terminated; }

  void Execute(void);
  void Execute(TCount numcmd);
  void ExecuteTo(TIndex line);

  TIndex ReadCommandRegister(void)
    { return CommandRegister; }
  xString ReadRegister(xString const index)
    { return RegisterManager[index].GetValue(); }
};




#endif
