/*
  RAM registers
  Copyright (C) 1999  Hardy Kahl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ----------------------------------------------------------------------------
//  Description      : RAM registers.
// ----------------------------------------------------------------------------




#include "register.h"
#include "exgen.h"




// xNamedRegister : -----------------------------------------------------------
xNamedRegister::xNamedRegister(xString name)
: Content("0"), Name(name)
{
}




xNamedRegister xNamedRegister::Set(xString const &value)
{
  Content = value;
  Optimize(Content);
  return *this;
}




xNamedRegister xNamedRegister::Set(xNamedRegister const &reg)
{
  Content = reg();
  return *this;
}




xNamedRegister xNamedRegister::Add(xNamedRegister const &reg)
{
  xNamedRegister result("0");
  xString        toadd = reg();
  char           carry = '0';

  if (~toadd > ~Content) {
    result.Content = toadd;
    toadd = Content;
  }
  else result.Content = Content;

  TIndex i = 0;

  while (i < ~toadd) {
    result.Content[i] = AddDigits(result.Content[i], toadd[i], carry);
    i++;
  }
  while ((i < ~result.Content) && (carry != '0')) {
    result.Content[i] = AddDigits(result.Content[i], '0', carry);
    i++;
  }
  if (carry != '0') result.Content += carry;

  return result;
}




xNamedRegister xNamedRegister::Sub(xNamedRegister const &reg)
{
  xNamedRegister result("0");
  xString        tosub = reg();
  char           carry = '0';
  TIndex         i;

  result.Content = Content;

  // check if result will be below zero
  if (~tosub > ~result.Content) {
    result.Content = "0";
    return result;
  }
  if (~tosub == ~result.Content) {
    i = ~result.Content - 1;
    while (tosub[i] == result.Content[i]) {
      if (i == 0) {
        result.Content = "0";
        return result;
      }
      i--;
    }
    if (tosub[i] > result.Content[i]) {
      result.Content = "0";
      return result;
    }
  }

  // substract ...
  i = 0;


  while (i < ~tosub) {
    result.Content[i] = SubDigits(result.Content[i], tosub[i], carry);
    i++;
  }
  while ((i < ~result.Content) && (carry != '0')) {
    result.Content[i] = SubDigits(result.Content[i], '0', carry);
    i++;
  }

  Optimize(result.Content);

  return result;
}




char xNamedRegister::AddDigits(char digit1, char digit2, char &carry)
{
  char sum = (digit1 - '0') + (digit2 - '0') + (carry - '0');
  if (sum >= 10) {
    carry = '1';
    sum -= 10;
  }
  else carry = '0';

  return '0' + sum;
}




char xNamedRegister::SubDigits(char digit1, char digit2, char &carry)
{
  digit1 -= '0';
  digit2 -= '0';
  digit2 += carry - '0';

  if (digit2 > digit1) {
    carry = '1';
    digit1 += 10;
  }
  else carry = '0';

  return '0' + digit1 - digit2;
}




void xNamedRegister::Optimize(xString &number)
{
  // delete all "left" zero digits
  while (number[~number - 1] == '0') {
    if (~number == 1) return;
    else number.DeleteRightThis();
  }
}




// xRegisterManager : ---------------------------------------------------------
void xRegisterList::Kill(xNamedRegister *&object)
{
  if (object) {
    delete object;
    object = 0;
  }
}




xNamedRegister &xRegisterManager::FindReg(xString const &index)
{
  xListIterator<xNamedRegister *> iterator(RegisterList);
  xNamedRegister *reg = 0;

  while (~iterator) {
    reg = iterator++;
    if (reg->GetName()==index) return *reg;
  }

  reg = new xNamedRegister(index);
  if (!(reg)) throw xMemoryException(ECMEM_GENERAL,sizeof(xNamedRegister));
  RegisterList += reg;

  return *reg;
}
