/*
  Extended string object
  Copyright (C) 1999  Hardy Kahl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ----------------------------------------------------------------------------
//  Description      : Extended string object
// ----------------------------------------------------------------------------
//  (c) Copyright 1996 by iXiONmedia, all rights reserved.
// ----------------------------------------------------------------------------




#ifdef __WINDOWS__
  #include <windows.h>
#endif
#include <string.h>
#include "xstring.h"
#include "exgen.h"




// Data objects --------------------------------------------------------------
xString numChars = XSTR_NUMCHARS;




// Case Conversion ------------------------------------------------------------
char chUpper(char c)
{
#ifdef __WINDOWS__
  return (char) AnsiUpper((LPSTR) c);
#else
  unsigned char uc = c;
  if ((c>='a') && (c<='z'))                     return c - 32;
  if (uc==132)                                  return 142;
  if (uc==148)                                  return 153;
  if (uc==129)                                  return 154;
  if ((uc==131) || (uc==133) || (uc==160))      return 'A';
  if ((uc==136) || (uc==138))                   return 'E';
  if (uc==130)                                  return 144;
  if ((uc==161) || (uc==141) || (uc==140))      return 'I';
  if ((uc==162) || (uc==149) || (uc==147))      return 'O';
  if ((uc==163) || (uc==151) || (uc==150))      return 'U';
  if (uc==164)                                  return 165;
  if (uc==151)                                  return 146;
  if (uc==135)                                  return 128;
  if (uc==134)                                  return 134;
  return c;
#endif
}




char chLower(char c)
{
#ifdef __WINDOWS__
  return (char) AnsiLower((LPSTR) c);
#else
  unsigned char uc = c;
  if ((c>='A') && (c<='Z'))     return  c + 32;
  if (uc==142)                  return 132;
  if (uc==153)                  return 148;
  if (uc==154)                  return 129;
  if (uc==144)                  return 130;
  if (uc==165)                  return 164;
  if (uc==146)                  return 145;
  if (uc==128)                  return 135;
  if (uc==143)                  return 134;
  return c;
#endif
}




// xString --------------------------------------------------------------------
// Static variables
TBool xString::FlagCaseSens = YES;




// Assignment and concatenation
xString &xString::operator=(xString const &source)
{
  SetLength(~source);
  if (~source) memcpy(!Buffer,!source.Buffer,Length);
  return *this;
}




xString &xString::operator=(char const *source)
{
  if (source) {
    SetLength(strlen(source));
    memcpy(!Buffer,source,Length);
  }
  else SetLength(0);
  return *this;
}




xString &xString::operator=(char const source)
{
  SetLength(1);
  Buffer[0] = source;
  return *this;
}




xString &xString::operator+=(xString const &source)
{
  if (~source) {
    TSize oldlen = Length;
    SetLength(Length+~source);
    memcpy(!Buffer+oldlen,!source.Buffer,~source);
  }
  return *this;
}




xString &xString::operator+=(char const *source)
{
  if (source) {
    TSize sourcelen = strlen(source);
    TSize oldlen = Length;
    SetLength(Length+sourcelen);
    memcpy(!Buffer+oldlen,source,sourcelen);
  }
  return *this;
}



xString &xString::operator+=(char const source)
{
  SetLength(Length+1);
  Buffer[Length-1] = source;
  return *this;
}



// Construction
xString &xString::MakeThis(TSize repeat,xString const &source)
{
  if (!repeat) return Clear();
  SetLength(~source*repeat);

  char *runner = !Buffer;
  while (repeat--) {
    memcpy(runner,!source.Buffer,~source);
    runner += ~source;
  }
  return *this;
}




xString &xString::MakeThis(TSize repeat,char const source)
{
  if (!repeat) return Clear();
  SetLength(repeat);
  memset(!Buffer,source,repeat);

  return *this;
}




xString xString::PercentBar(TSize count,TSize total,TSize length,
  char const full,char const empty)
{
  xString temp;
  while (count > 10 * length) {
    count >>= 1;
    total >>= 1;
  }
  TSize fullchars=count * length / total;
  temp = Make(fullchars,full)+Make(length-fullchars,empty);
  return temp;
}



// Comparison
TBool xString::operator==(xString const &operand) const
{
  if (Length != ~operand) return NO;
  if (FlagCaseSens) return memcmp(!Buffer,!operand.Buffer,NUM_MIN(Length,~operand)) == 0;
  else {
    xString tempop = operand.Upper(),
            tempthis = Upper();
    return memcmp(!tempthis.Buffer,!tempop.Buffer,NUM_MIN(Length,~operand)) == 0;
  }
}




TBool xString::operator<(xString const &operand) const
{
  if (FlagCaseSens) {
    int result = memcmp(!Buffer,!operand.Buffer,NUM_MIN(Length,~operand));
    if (result == 0) return Length<~operand;
    else return  result < 0;
  }
  else {
    xString tempop = operand.Upper(),
            tempthis = Upper();
    int result = memcmp(!tempthis.Buffer,!tempop.Buffer,NUM_MIN(Length,~operand));
    if (result == 0) return Length<~operand;
    else return  result < 0;
  }
}



// Extraction
char &xString::operator[](TIndex position) const
{
  if (position >= Length) EX_THROW(Generic,EC_INDEX)
  return Buffer[position];
}




xString xString::operator()(TDelta start) const
{
  if (start) {
    if (start<0) start = Length + start;
    if (start<0) { EX_THROW(Generic,EC_INDEX) }
    if (start>(TDelta)Length) { EX_THROW(Generic,EC_INDEX) }
    xString temp;
    temp.SetLength(Length-start);
    memcpy(!temp.Buffer,!Buffer+start,temp.Length);
    return temp;
  }
  else return *this;
}




xString xString::operator()(TDelta start,TSize length) const
{
  xString temp = operator()(start).Clip(length);
  if (~temp < length) EX_THROW(Generic,EC_INVALIDPAR)
  return temp;
}




// Insertion and Deletion
xString xString::Insert(TIndex position,xString const &substr) const
{
  if (position > Length) EX_THROW(Generic,EC_INDEX)
  return operator()(0,position)+substr+operator()(position);
}




xString xString::Overwrite(TIndex position,xString const &substr) const
{
  if (position > Length) EX_THROW(Generic,EC_INDEX)
  if (!~substr) return *this;
  xString temp;
  if (position + ~substr >= Length) return operator()(0,position)+substr;
  else return operator()(0,position)+substr+operator()(position+~substr);
}




xString xString::Delete(TIndex start,TSize size) const
{
  xString temp;
  if (start) temp = operator()(0,start);
  if (Length > start+size) temp += operator>>(Length - (start+size));
  else if (Length < start+size) EX_THROW(Generic,EC_INVALIDPAR);
  return temp;
}




xString xString::DeleteChar(char ch) const
{
  xString temp;
  for (TIndex i=0; i<Length; i++) {
    if (Buffer[i] != ch) temp += Buffer[i];
  }
  return temp;
}




xString xString::DeleteRight(TSize size) const
{
  if (size > Length) EX_THROW(Generic,EC_INVALIDPAR)
  return Clip(Length-size);
}




// Tools
xString &xString::ClipThis(TSize maxlength)
{
  if (maxlength >= Length) return *this;
  SetLength(maxlength);
  return *this;
}




xString xString::Clip(TSize maxlength) const
{
  xString temp = *this;
  temp.ClipThis(maxlength);
  return temp;
}




xString xString::Limit(TSize length,char filler) const
{
  if (Length < length) return *this + Make(length-Length,filler);
  else return Clip(length);
}




xString xString::PreLimit(TSize length,char filler) const
{
  if (Length < length) return Make(length-Length,filler) + *this;
  else return Clip(length);
}




xString xString::Reverse() const
{
  xString temp;
  temp.SetLength(Length);

  for (TIndex index = 0;index < Length;index++)
    temp.Buffer[index] = Buffer[Length-index-1];
  return temp;
}




xString xString::Optimize() const
{
  TSize countspaces = 0;
  while ((countspaces < Length) && Buffer[countspaces] == ' ') countspaces++;
  xString temp = Delete(0,countspaces);

  if (~temp) {
    countspaces = 0;
    while ((countspaces < ~temp) && temp[~temp-countspaces-1] == ' ')
      countspaces++;
    temp = temp.DeleteRight(countspaces);
  }
  return temp;
}




TBool xString::InSet(xString const &set) const
{
  for (TIndex index = 0;index < Length;index++)
    if (!set.CountChar(Buffer[index])) return NO;
  return YES;
}




// Finding
TSize xString::CountChar(char const ch) const
{
  TSize count = 0;
  for (TIndex index = 0;index < Length;index++)
    if (Buffer[index]==ch) count++;

  return count;
}




TIndex xString::Locate(xString const &searchstr,TIndex startat) const
{
  if (startat >= Length) EX_THROW(Generic,EC_INDEX);
  if (!~searchstr) EX_THROW(Generic,EC_INVALIDPAR);
  if (~searchstr>Length) EX_THROW(Generic,EC_ITEMNOTFOUND);

  TIndex comppos = startat;
  while (comppos <= Length-~searchstr) {
    TIndex index = 0;
    while ((operator[](comppos+index) == searchstr[index]) && (++index < ~searchstr)) ;
    if (index == ~searchstr) { return comppos; }
    comppos++;
  }
  EX_THROW(Generic,EC_ITEMNOTFOUND)
}




TIndex xString::Locate(char const searchchar,TIndex startat) const
{
  if (startat >= Length) EX_THROW(Generic,EC_INDEX);
  if (!Length) EX_THROW(Generic,EC_ITEMNOTFOUND);

  TIndex comppos = startat;
  while (comppos < Length) {
    if (operator[](comppos) == searchchar) { return comppos; }
    comppos++;
  }
  EX_THROW(Generic,EC_ITEMNOTFOUND)
}




TIndex xString::LocateLast(xString const &searchstr) const
{
  if (!~searchstr) EX_THROW(Generic,EC_INVALIDPAR);
  if (~searchstr>Length) EX_THROW(Generic,EC_ITEMNOTFOUND);

  TIndex comppos = Length-~searchstr;
  do {
    TIndex index = 0;
    while ((operator[](comppos+index) == searchstr[index]) && (++index < ~searchstr)) ;
    if (index == ~searchstr) { return comppos; }
    comppos--;
  } while (comppos--);
  EX_THROW(Generic,EC_ITEMNOTFOUND)
}




TIndex xString::LocateLast(char const searchchar) const
{
  if (!Length) EX_THROW(Generic,EC_ITEMNOTFOUND);

  TIndex comppos = Length-1;
  do {
    if (operator[](comppos) == searchchar) { return comppos; }
  } while (comppos--);
  EX_THROW(Generic,EC_ITEMNOTFOUND)
}




xString xString::Replace(xString const &source,xString const &destination) const
{
  xString temp = *this;
  try {
    while (1) {
      TIndex index = temp.Locate(source);
      temp.DeleteThis(index,~source);
      temp.InsertThis(index,destination);
    }
  }
  EX_CATCHCODE(Generic,EC_ITEMNOTFOUND,)
  return temp;
}




// Case conversion
xString &xString::LowerThis()
{
  for (TIndex index = 0;index < Length;index++)
    Buffer[index] = chLower(Buffer[index]);

  return *this;
}




xString &xString::UpperThis()
{
  for (TIndex index = 0;index < Length;index++)
    Buffer[index] = chUpper(Buffer[index]);

  return *this;
}




// Numeric conversion
xString xString::Unsigned2Base(unsigned long value,char digits,char radix)
{
  xString temp;
  do {
    temp = numChars[value % radix]+temp;
    value /= radix;
    if (digits) digits--;
  } while (value || digits);
  return temp;
}




xString xString::Signed2Base(signed long value,char digits,char radix)
{
  if (value < 0) return "-"+Unsigned2Base(-value,digits,radix);
  else return Unsigned2Base(-value,digits,radix);
}




xString xString::Bytes2Dec(TSizeD bytes)
{
  if (bytes>(TSizeD) 10*1024*1024)
    return Unsigned2Dec(bytes / ((TSizeD) 1024*1024))+" MB";
  if (bytes>(TSizeD) 10*1024)
    return Unsigned2Dec(bytes / ((TSizeD) 1024))+" KB";
  return Unsigned2Dec(bytes)+" Byte";
}




unsigned long xString::EvalUnsigned() const
{
  if (Length == 0) return 0;
  xString numstr = (*this).Upper();

  if ((numstr << 2) == "0X")
    return Evaluate(numstr(2),0x10);
  if ((numstr << 1) == "$")
    return Evaluate(numstr(1),0x10);

  char lastchar = (numstr >> 1)[0];
  if (lastchar == 'H') return Evaluate(numstr.DeleteRight(1),0x10);
  if (lastchar == 'B') return Evaluate(numstr.DeleteRight(1),2);
  if (lastchar == 'D') return Evaluate(numstr.DeleteRight(1),10);
  if (lastchar == 'O') return Evaluate(numstr.DeleteRight(1),8);

  return Evaluate(numstr);
}




signed long xString::EvalSigned() const
{
  if (Length == 0) return 0;
  if (Buffer[0] == '-')
    return -operator()(1).EvalUnsigned();
  else
    return EvalUnsigned();
}




// Protected members
unsigned long xString::Evaluate(xString const &numstr,char radix) const
{
  if (~numstr == 0) return 0;
  unsigned long value = 0, mulvalue = 1;
  TIndex index = ~numstr-1;

  do {
    unsigned long digvalue;
    try {
      digvalue = numChars.Locate(numstr[index]);
    }
    catch (...) { EX_THROWINFO(Generic,EC_CANNOTEVALUATE,numstr) }
    value += mulvalue * digvalue;
    mulvalue *= radix;
  } while (index--);

  return value;
}




// friends --------------------------------------------------------------------
xString operator+(char const *s1,xString const &s2)
{
  xString temp = s1;
  return temp + s2;
}




xString operator+(char const s1,xString const &s2)
{
  xString temp = s1;
  return temp + s2;
}
