/*
  plic compiler - second pass
  Copyright (C) 1999  Achim Kern

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ---------------------------------------------------------------------------- 
//  Description      : Implementation of the second pass of the pascalli comp.
// ---------------------------------------------------------------------------- 
//  Creation	     : 06.07.1999
// ---------------------------------------------------------------------------- 
//  Remarks          : 
// ---------------------------------------------------------------------------- 
//  Modifications    : 
//  	$Log: backend_2.cpp,v $
// Revision 1.2  1999/07/19  14:50:41  thieleek
// first complete revision
//
// Revision 1.1.1.1  1999/07/13  13:35:49  thieleek
// sources Imported
//
// ---------------------------------------------------------------------------- 





#include "plic_Backend2.h"
#include "plic_Copier.h"
#include <fstream.h>
#include <iostream.h>




xPascalliVisitor_secondpass::xPascalliVisitor_secondpass(TIndex lastcommand)
{
  //  cout << "second pass started" << endl;
  LastCommand = lastcommand;
  Status = OK;
};




xPascalliVisitor_secondpass::~xPascalliVisitor_secondpass()
{
};




xNode* xPascalliVisitor_secondpass::
startCompilation(xNode * topnode)
{
  TopNode = (xProgramNode*)(topnode);
  return xPascalliTraverser::startCompilation(topnode);
};




//
// to translate:
//   function f1(arg1,...,argn)
//   begin
//     f1 := g(1, arg1, ..., argn, 0, ..., 0);
//   end
//
//   function g(funcsel,arg1,...,argk) 
//   begin
//     s1;
//   end
//
// translation:
//   function f1(_p1,...,_pn)
//   begin
//     CurrentIncarnation := 1;
//     NextIncarnation := 2;
//     funcsel := 1;
//     p1[1] := arg1;
//     ...
//     pn[1] := argn;
//       s;
//     f1 := g[1];
//   end
// 
// Remarks:
//   - it is not necessary to let a memorymanager handle
//     the addresses for arg1,..,argn because they do not
//     appear in any other place in the program.
//     Because they have been deleted by the first pass
//

void xPascalliVisitor_secondpass::
visit_xProgramNode(xProgramNode & node)
{
  xFunction*    NewFunction = 0;
  xWhile*       OuterLoop = 0;
  xIfStatement* IncControl = 0;

  NewFunction = new xFunction(new xIdentifier("f"), *(new xIdentifierList), new xCommandBlock);

  // insert parameters for the new function
  for (TIndex ArgumentNr = 0; ArgumentNr < ~(node.Functions[0]->Arguments); ArgumentNr++)
  {
     NewFunction->Arguments += new xIdentifier("_p" + xString::Unsigned2Dec(ArgumentNr));
  };

  // do initaliziation of important variables
  // _One = 1;
  // _One is used as a variable containing 1. This is only for effizent reason. Having a fix variable containing 1 enables us just to
  // use the name of these variable in every case we need a 1
  ((xCommandBlock*)(NewFunction->Command))->addCommand(new xAssignment(new xVariable("_One"), new xConstant("1")));
  // CurrentIncarnation = 1
  ((xCommandBlock*)(NewFunction->Command))->addCommand(new xAssignment(new xVariable("_CurrentInc"), new xVariable("_One")));
  // NextIncarnation = 2
  ((xCommandBlock*)(NewFunction->Command))->addCommand(new xAssignment(new xVariable("_NextInc"), new xConstant("2")));
  // FuncSel = 1
  ((xCommandBlock*)(NewFunction->Command))->addCommand(new xAssignment(new xArray("_FuncSel", new xVariable("_CurrentInc")), 
								       new xVariable("_One")));

  // add assignments for the parameter of the new function
  // These assignments do have the form formerparamter[1]:= newparameter, wher formerparamter is the paramtername of the corresponding
  // parameter of the first function and newparamter is something like _pxx
  for (TIndex Parameter = 0; Parameter < ~(node.Functions[0]->Arguments); Parameter++)
  {
    xString FormerParameter = node.Functions[0]->Arguments[Parameter]->Name;
    xString NewParameter = "_p" + xString::Unsigned2Dec(Parameter);
    ((xCommandBlock*)(NewFunction->Command))->addCommand(new xAssignment(new xArray(FormerParameter, new xVariable("_One")), new xVariable(NewParameter)));
  };
  
  OuterLoop = new xWhile(new xGreater(new xVariable("_CurrentInc"), new xConstant("0")), new xCommandBlock);
  // the ReturnAddress is zero if we are in a new incarnation. Otherwise it contains the command number of the command being executed next
  ((xCommandBlock*)(OuterLoop->Command))->addCommand(new xAssignment(new xIdentifier("_NextCommand"), new xArray("_ReturnAddress", new xIdentifier("_CurrentInc"))));
  
  // prepare translation of the several functions
  InsertationPoint.push((xCommandBlock*)(OuterLoop->Command));
  MemManager.init(node.Functions);
  // translate the functions
  for (TIndex Counter = 0; Counter < ~(node.Functions); Counter++)
  {
    node.Functions[Counter]->AcceptVisitor(*this);
  };

  // Now insert the control for the several incarnation possible to the function _f
  // if we perform a jump than next command # 0. If nextCommand = 0 then we are finished with one incarnation
  // if (NextCommand = 0 then... 
  IncControl = new xIfStatement(new xEqual(new xVariable("_NextCommand"), new xConstant("0")), new xCommandBlock, new xCommandBlock);
  ((xCommandBlock*)(IncControl->IfTwig))->addCommand(new xAssignment(new xIdentifier("_CurrentInc"), new xArray("_InvocingInc", new xIdentifier("_CurrentInc"))));
  ((xCommandBlock*)(IncControl->ElseTwig))->addCommand(new xAssignment(new xVariable("_CurrentInc"), new xVariable("_NextInc")));
  ((xCommandBlock*)(IncControl->ElseTwig))->addCommand(new xAssignment(new xVariable("_NextInc"), new xAddition(new xVariable("_NextInc"), new xVariable("_One"))));
  ((xCommandBlock*)(OuterLoop->Command))->addCommand(IncControl);
  
  ((xCommandBlock*)(NewFunction->Command))->addCommand(OuterLoop);
  
  // assign the value of the first function to the new function
  ((xCommandBlock*)(NewFunction->Command))->addCommand(new xAssignment(new xVariable("_f"), new xArray(node.Functions[0]->Name.Name, new xVariable("_One"))));

  Results.push(NewFunction);
};




void xPascalliVisitor_secondpass::
visit_xFunction(xFunction & node)
{
  xCommandBlock* CurrentCommandBlock = 0;
  xCommandBlock* IfCommand = 0;
  xEqual*        Condition = 0;

  CurrentCommandBlock = InsertationPoint.pop();
  // if (FuncSel = nummer der aktuellen Funktion) then ...
  IfCommand = new xCommandBlock;
  Condition = new xEqual(new xArray("_FuncSel", new xVariable("_CurrentInc")), 
			 new xConstant(xString::Unsigned2Dec(MemManager.getFunctionnumber(node.Name.Name))));
  CurrentCommandBlock->addCommand(new xIfStatement(Condition, IfCommand, 0));
 
  InsertationPoint.push(IfCommand);
  node.Command->AcceptVisitor(*this);
  // pop IfCommand CommandBlock from stack
  InsertationPoint.pop();
  // add the result of the traversation to IfCommand
  IfCommand->addCommand((xCommand*)(Results.pop()));
  // push the top mainprogramm CommandBlock back on the stack
  InsertationPoint.push(CurrentCommandBlock); 
 
  //  Results.push(&node);
};




void xPascalliVisitor_secondpass::
visit_xAssignment(xAssignment & node)
{
  //  xCommandBlock* CurrentCommandBlock = 0;
  xIfStatement* IfCommand = 0;
  xString       CommandNr;
  xIdentifier*  Dest = 0;
  xExpression*  Source = 0;

   node.Destination->AcceptVisitor(*this);
   node.Source->AcceptVisitor(*this);
   Source = (xExpression*)(Results.pop());
   Dest = (xIdentifier*)(Results.pop());
  
   CommandNr = xString::Unsigned2Dec(node.Alpha);
   IfCommand = new xIfStatement(new xSmallerEqual(new xIdentifier("_NextCommand"), new xConstant(CommandNr)), new xCommandBlock, 0);
   ((xCommandBlock*)(IfCommand->IfTwig))->addCommand(new xAssignment(Dest, Source));
  
   Results.push(IfCommand);
};



// to translate:
//  while(condition) do command end;
//
// Translation:
//  if (next command  <= beta) then 
//  begin
//    if (next command >= alpha) then command';
//    while (condition AND (next command <= beta)) do command';
//  end
//
// Implementation:
//  visit Command and Condition
//  the results of this traversation are pushed to the Results stack
//  insert the new nodes as disscussed above
//  
void xPascalliVisitor_secondpass::
visit_xWhile(xWhile & node)
{
  xCommandBlock*  Program = 0;
  xCommandBlock*  IfCommand = 0;
  xPascalliCopier CopyMan;
    
  Program = new xCommandBlock;
  IfCommand = new xCommandBlock;

  node.Command->AcceptVisitor(*this);
  node.Condition->AcceptVisitor(*this);

  node.Condition = new xAnd((xCondition*)(Results.pop()), 
			    new xSmallerEqual(new xIdentifier("_NextCommand"), new xConstant(xString::Unsigned2Dec(node.Beta))));
  node.Command = (xCommand*)(Results.pop());

  // if (_NextCommand <= node.beta) then ...
  Program->addCommand(new xIfStatement
		      (new xSmallerEqual
		       (new xIdentifier("_NextCommand"), new xConstant(xString::Unsigned2Dec(node.Beta))),
		      IfCommand, 0));
  // if (_NextCommand >= node.alpha) then node.Command
  IfCommand->addCommand(new xIfStatement
			(new xGreaterEqual(new xIdentifier("_NextCommand"), new xConstant(xString::Unsigned2Dec(node.Alpha))),
			(xCommand*)(CopyMan.startCompilation(node.Command)), 0));
  // copy node.command and proced the while loop
  IfCommand->addCommand(&node);

  Results.push(Program);
};




void xPascalliVisitor_secondpass::
visit_xIdentifier(xIdentifier & node)
{
   if (node.Name[0] == '_')
   {
     Results.push(&node);
   }
   else
   {
     Results.push(new xArray(node.Name, new xIdentifier("_CurrentInc")));
     delete (&node);
   };
};



//
// we look at arrays as if they were 2 dimensional: a[Index, Incarnation]
// than we flatten these array into 1 dimensional ones by calculating their
// index by a diagonal strategie;
// a[Index, Incarnation] => a[(((Index * (Index + 1))/2) + Incarnation)]
//
void xPascalliVisitor_secondpass::
visit_xArray(xArray & node)
{
  xCommandBlock* Program = 0;
  xVariable*   Index = 0;
  
  node.Index->AcceptVisitor(*this);
  Index = MemManager.New();

  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);
  
  Program->addCommand(new xAssignment(Index, (xExpression*)(Results.pop())));
  node.Index =new xAddition(new xDivision(new xMultiplication(new xIdentifier(Index->Name), new xAddition(new xIdentifier(Index->Name), new xIdentifier("_One"))), new xConstant("2")), new xIdentifier("_CurrentInc"));
  
  Results.push(&node);
};




void xPascalliVisitor_secondpass::
visit_xFunctionCall(xFunctionCall & node)
{
  xCommandBlock* Program = 0;
  xCommandBlock* JumpBlock = 0;
  xCommandBlock* ReturnBlock = 0;
  xFunction*     CurrentFunction = 0;
  xVariable*     Result;

  JumpBlock = new xCommandBlock;
  // set retrunaddress
  // ReturnAddress[_CurrentIncarnation]=node.alpha
  JumpBlock->addCommand(new xAssignment
			(new xArray("_ReturnAddress", new xVariable("_CurrentInc")), 
			 new xConstant(xString::Unsigned2Dec(node.Alpha))));
  // set incarantions, prepare next incarnation
  JumpBlock->addCommand(new xAssignment(new xArray("_InvocedInc", new xVariable("_CurrentInc")), new xVariable("_NextInc")));
  JumpBlock->addCommand(new xAssignment(new xArray("_InvocingInc", new xVariable("_NextInc")), new xVariable("_CurrentInc")));
  // FuncSel = number of function
  JumpBlock->addCommand(new xAssignment
			(new xArray("_FuncSel", new xVariable("_NextInc")), 
			 new xConstant(xString::Unsigned2Dec(MemManager.getFunctionnumber(node.Name)))));
  // set NextCommand to a number greater than the last command of the program.
  // This is done to ensure that no command is executed before the end of the program
  JumpBlock->addCommand(new xAssignment
			(new xVariable("_NextCommand"), 
			 new xConstant(xString::Unsigned2Dec(LastCommand))));
  // assign the parametervalues
  CurrentFunction = TopNode->Functions[MemManager.getFunctionnumber(node.Name)-1];
  for (TIndex Number = 0; Number < ~(node.Arguments); Number++)
  {
    node.Arguments[Number]->AcceptVisitor(*this);
    JumpBlock->addCommand(new xAssignment
			  (new xArray(CurrentFunction->Arguments[Number]->Name, new xVariable("_NextInc")),
			  (xExpression*)(Results.pop())));
  };

  // the next paragraph is executed when we return for a functioncall
  ReturnBlock = new xCommandBlock;
  Result = MemManager.New();
  ReturnBlock->addCommand(new xAssignment
			  (new xVariable(Result->Name), 
			   new xArray(CurrentFunction->Name.Name, new xArray("_InvocedInc", new xVariable("_CurrentInc")))));
  ReturnBlock->addCommand(new xAssignment(new xVariable("_NextCommand"), new xConstant("0")));

  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);
  Program->addCommand(new xIfStatement
		      (new xSmaller(new xVariable("_NextCommand"), new xConstant(xString::Unsigned2Dec(node.Alpha))), 
		       JumpBlock, 
		       0));
  Program->addCommand(new xIfStatement
		      (new xEqual(new xVariable("_NextCommand"), new xConstant(xString::Unsigned2Dec(node.Alpha))), 
		       ReturnBlock, 
		       0));

  
  Results.push(new xVariable(Result->Name));
};








































































