/*
  plic compiler - third pass
  Copyright (C) 1999  Achim Kern

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ---------------------------------------------------------------------------- 
//  Description      : Implementation of the first pass of the Pascalli to
//                     Pascallino compiler
// ---------------------------------------------------------------------------- 
//  Creation         : 10.6.1999
// ---------------------------------------------------------------------------- 
//  Remarks          : One is a Variable containing the value 1.
//                     A corresponding line is added to the program in 
//                     xPascalliVisitor::visit_xProgramNode.
// ---------------------------------------------------------------------------- 
//  Modifications    : 
// ---------------------------------------------------------------------------- 




#include "plic_Backend3.h"
#include "plic_Copier.h"
#include <fstream.h>
#include <iostream.h>




xPascalliVisitor_thirdpass::xPascalliVisitor_thirdpass()
{
  //  cout << "third pass started" << endl;
};




xPascalliVisitor_thirdpass::~xPascalliVisitor_thirdpass()
{
};





void xPascalliVisitor_thirdpass::
visit_xAssignment(xAssignment & node)
{
  Side = Left;
  node.Destination->AcceptVisitor(*this);
  node.Destination = (xIdentifier*)(Results.pop());
  Side = Right;
  node.Source->AcceptVisitor(*this);
  node.Source = (xExpression*)(Results.pop());

  Results.push(&node);
};




void xPascalliVisitor_thirdpass::
visit_xWhile(xWhile & node)
{
  xCommandBlock*  WhileCommand = 0;
  xExpression*    ConditionCopy = 0;
  xVariable*      Condition = 0;
  xPascalliCopier Copier;

  ConditionCopy = (xExpression*)(Copier.startCompilation(node.Condition));
  node.Condition->AcceptVisitor(*this);
  Condition = (xVariable*)(Results.pop());
  WhileCommand = new xCommandBlock;
  InsertationPoint.push(WhileCommand);
  node.Command->AcceptVisitor(*this);
  WhileCommand->addCommand((xCommand*)(Results.pop()));
 // calc the condition a seccond time at the end of the while loop
  ConditionCopy->AcceptVisitor(*this);
  ConditionCopy = (xExpression*)(Results.pop());
  WhileCommand->addCommand(new xAssignment(new xVariable(Condition->Name), ConditionCopy));
  
  InsertationPoint.pop();

  node.Command = WhileCommand;
  node.Condition = new xUnEqual(Condition, new xConstant("0"));

  Results.push(&node);
};




//
// to translate:
//   if cond then s1 else s2
//
// translation:
//   DoLoop := Value(cond);
//   DoElseLoop := 1;
//   while(DoLoop # 0) do
//   begin
//     s1;
//     DoLoop := 0;
//     DoElseLoop := 0;
//   end
//   if there is an ElseTwig we continue:
//   while(DoElseLoop # 0) do
//   begin
//     s2;
//     DoElseLoop := 0;
//   end
//
// steps of translation:
//   1) translate cond;
//   2) assign cond->ResultAddress to DoLoop;
//   3) create the Program;
//

void xPascalliVisitor_thirdpass::
visit_xIfStatement(xIfStatement & node)
{
  xCommandBlock* IfTwig = 0;
  xCommandBlock* ElseTwig = 0;
  xVariable*     DoLoop = 0;
  xVariable*     DoElseLoop = 0;
  xCommandBlock* Program = 0;

  //  Program = InsertationPoint.pop();
  // InsertationPoint.push(Program);
  Program = new xCommandBlock;

  node.Condition->AcceptVisitor(*this);
  InsertationPoint.push(new xCommandBlock);
  node.IfTwig->AcceptVisitor(*this);
  IfTwig = InsertationPoint.pop();
  IfTwig->addCommand((xCommand*)(Results.pop()));

  DoLoop = MemManager.New();
  Program->addCommand(new xAssignment(DoLoop, (xExpression*)(Results.pop())));
  // add to while commandblock:
  // DoLoop = 0
  IfTwig->addCommand(new xAssignment(new xVariable(DoLoop->Name), new xConstant("0")));
  if (node.ElseTwig != 0)
  {
    DoElseLoop = MemManager.New();
    Program->addCommand(new xAssignment(DoElseLoop, new xVariable("One")));
    // add to while commandblock:
    // DoElseLoop = 0
    IfTwig->addCommand(new xAssignment(new xVariable(DoElseLoop->Name), new xConstant("0")));
  };
  Program->addCommand(new xWhile(new xUnEqual(new xVariable(DoLoop->Name), new xConstant("0")), IfTwig));
  
  if (node.ElseTwig != 0)
  {
    InsertationPoint.push(new xCommandBlock);
    node.ElseTwig->AcceptVisitor(*this);
    ElseTwig = InsertationPoint.pop();
    ElseTwig->addCommand((xCommand*)(Results.pop()));
    ElseTwig->addCommand(new xAssignment(new xVariable(DoElseLoop->Name), new xConstant("0")));
    Program->addCommand(new xWhile(new xUnEqual(new xVariable(DoElseLoop->Name), new xConstant("0")), ElseTwig));
  };

  Results.push(Program);
};




//
// to translate:
//   a and b
//
// translation:
//   tmp1 := 1 - value(a);
//   tmp2 := 1 - value(b);
//   tmp1 := tmp1 + tmp2;
//   tmp1 := 1 - tmp1;
//
// steps of translation:
//   1) use the law of Demorgan to create: not(not(a) or not(b));
//   2) not(a): 1-value(a)
//   3) not(a) or not(b): (1-value(a)) + (1-value(b));
//   4) not (not(a) or not(b)): 1-((1-value(a)) + (1-value(b)));
//
// Remarks:
//   The braces are absolutly necessary because of the modified difference of the ram these expression are no longer commutativ
//

void xPascalliVisitor_thirdpass::
visit_xAnd(xAnd & node)
{
  xVariable*     Tmp1 = 0;
  xVariable*     Tmp2 = 0;
  xCommandBlock* Program = 0;

  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);

  Side=Right;
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);

  Tmp1 = MemManager.New();
  Tmp2 = MemManager.New();
  
  Program->addCommand(new xAssignment(Tmp1, new xSubtraction(new xVariable("One"), (xExpression*)(Results.pop()))));
  Program->addCommand(new xAssignment(Tmp2, new xSubtraction(new xVariable("One"), (xExpression*)(Results.pop()))));
  Program->addCommand(new xAssignment(new xVariable(Tmp1->Name), new xAddition(new xVariable(Tmp1->Name), new xVariable(Tmp2->Name))));
  Program->addCommand(new xAssignment(new xVariable(Tmp1->Name), new xSubtraction(new xVariable("One"), new xVariable(Tmp1->Name))));

  Results.push(new xVariable(Tmp1->Name));
};




//
// to translate:
//   a or b
//
// translation:
//   tmp:=value(a) + value(b);
//

void xPascalliVisitor_thirdpass::
visit_xOr(xOr & node)
{
  xVariable*     Tmp1 = 0;
  xCommandBlock* Program = 0;

  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);

  Side=Right;
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);

  Tmp1 = MemManager.New();
  
  Program->addCommand(new xAssignment(Tmp1, new xAddition((xExpression*)(Results.pop()), (xExpression*)(Results.pop()))));

  Results.push(new xVariable(Tmp1->Name));
};




//
// to translate:
//   not(a);
//
// translation:
//   tmp:= 1 - value(a);
//

void xPascalliVisitor_thirdpass::
visit_xNot(xNot & node)
{
  xVariable*     Tmp1 = 0;
  xCommandBlock* Program = 0;

  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);

  Side=Right;
  node.Expression->AcceptVisitor(*this);

  Tmp1 = MemManager.New();
  
  Program->addCommand(new xAssignment(Tmp1, new xSubtraction(new xVariable("One"), (xExpression*)(Results.pop()))));

  Results.push(new xVariable(Tmp1->Name));
};





//
// to translate:
//   (b >= a)
//
// this is not translated directly.
//
// steps of translation:
//   1) convert (b >= a) into (a <= b);
//   2) call visit_xSmallerEqual to translate (a <= b);
// 

void xPascalliVisitor_thirdpass::
visit_xGreaterEqual(xGreaterEqual & node)
{
  visit_xSmallerEqual(*(new xSmallerEqual(node.Right, node.Left)));
  node.Right = 0;
  node.Left = 0;
  delete(&node);
};




//
// to translate:
//   (b > a)
//
// this is not translated directly.
//
// steps of translation:
//   1) convert (b > a) into (a < b);
//   2) call visit_xSmaller to translate (a < b);
// 

void xPascalliVisitor_thirdpass::
visit_xGreater(xGreater & node)
{
  visit_xSmaller(*(new xSmaller(node.Right, node.Left)));
  node.Left = 0;
  node.Right = 0;
  delete(&node);
};




//
// to translate:
//   (a = b);
//
// translation:
//   tmp1 := value(a) - value(b);
//   tmp2 := value(b) - value(a);
//   tmp1 := tmp1 + tmp2;
//   tmp1 := 1 - tmp1;
//
// steps of translation:
//   1) (a=b) is equivalent to not(a<b) and not(b<a);
//   2) law of Demorgan: not( (a<b) or (b<a) );
//   3) not( (a<b) or (b<a) ): 1-( (value(a) - value(b)) + (value(b) - value(a)) );
//

void xPascalliVisitor_thirdpass::
visit_xEqual(xEqual & node)
{
  xVariable*     Tmp1;
  xVariable*     Tmp2;
  xVariable*   LeftTerm;
  xVariable*   RightTerm;
  xCommandBlock* Program;

  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);

  Side=Right;
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  
  RightTerm = (xVariable*)(Results.pop());
  LeftTerm = (xVariable*)(Results.pop());
  Tmp1 = MemManager.New();
  Tmp2 = MemManager.New();
  
  Program->addCommand(new xAssignment(Tmp1, new xSubtraction(LeftTerm, RightTerm)));
  Program->addCommand(new xAssignment(Tmp2, new xSubtraction(new xVariable(RightTerm->Name), new xVariable(LeftTerm->Name))));
  Program->addCommand(new xAssignment(new xVariable(Tmp1->Name), new xAddition(new xVariable(Tmp1->Name), new xVariable(Tmp2->Name))));
  Program->addCommand(new xAssignment(new xVariable(Tmp1->Name), new xSubtraction(new xVariable("One"), new xVariable(Tmp1->Name))));
  
  Results.push(new xVariable(Tmp1->Name));
};




//
// to translate:
//   (a<b)
//
// translation:
//   tmp := value(b)-value(a);
//

void xPascalliVisitor_thirdpass::
visit_xSmaller(xSmaller & node)
{
  xVariable* Tmp = 0;
  xCommandBlock* Program = 0;
  xExpression* RightPart = 0;
  xExpression* LeftPart = 0;

  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);

  Side = Right;
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  RightPart = (xExpression*)(Results.pop());
  LeftPart = (xExpression*)(Results.pop());

  Tmp = MemManager.New();
  Program->addCommand(new xAssignment(Tmp, new xSubtraction(RightPart, LeftPart)));
  
  Results.push(new xVariable(Tmp->Name));
};




//
// to translate:
//   (a<=b) 
// 
// this is not translated directly
//
// steps of translation:
//   1) convert (a<=b) into not(b<a);
//   2) call visit_xSmaller( (b<a) );
//   3) add tmp := One - Newnode->ResultAddress;
//

void xPascalliVisitor_thirdpass::
visit_xSmallerEqual(xSmallerEqual & node)
{
  xVariable*     Tmp;
  xCommandBlock* Program;
  
  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);

  visit_xSmaller(*(new xSmaller(node.Right, node.Left)));
  
  Tmp = MemManager.New();
  Program->addCommand(new xAssignment(Tmp, new xSubtraction(new xVariable("One"), (xExpression*)(Results.pop()))));

  Results.push(new xIdentifier(Tmp->Name));
};




//
// to translate:
//   (a # b) 
// 
// this is not translated directly
//
// steps of translation:
//   1) convert (a # b) into not(a = b);
//   2) call visit_xSmaller( (a = b) );
//   3) add  Newnode->ResultAddress := One - Newnode->ResultAddress;
//

void xPascalliVisitor_thirdpass::
visit_xUnEqual(xUnEqual & node)
{
  xCommandBlock* Program = 0;
  xVariable*     Tmp = 0;

  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);

  visit_xEqual(*(new xEqual(node.Left, node.Right)));
  Tmp = MemManager.New();
  Program->addCommand(new xAssignment(Tmp, new xSubtraction(new xVariable("One"), (xExpression*)(Results.pop()))));

  Results.push(new xVariable(Tmp->Name));
};




void xPascalliVisitor_thirdpass::
visit_xAddition(xAddition & node)
{
  xCommandBlock* Program = 0;
  xVariable*     Tmp = 0;
  xExpression*   Left = 0;
  xExpression*   Right = 0;

  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);

  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  Right = (xExpression*)(Results.pop());
  Left = (xExpression*)(Results.pop());

  Tmp = MemManager.New();
  Program->addCommand(new xAssignment(Tmp, new xAddition(Left, Right)));

  Results.push(new xVariable(Tmp->Name));
};




void xPascalliVisitor_thirdpass::
visit_xSubtraction(xSubtraction & node)
{
  xCommandBlock* Program = 0;
  xVariable*     Tmp = 0;
  xExpression*   Left = 0;
  xExpression*   Right = 0;

  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);

  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  Right = (xExpression*)(Results.pop());
  Left = (xExpression*)(Results.pop());

  Tmp = MemManager.New();
  Program->addCommand(new xAssignment(Tmp, new xSubtraction(Left, Right)));

  Results.push(new xVariable(Tmp->Name));
};




//
// to translate:
//   a * b
//
// translation:
//   Counter := value(b);
//   while (Counter # 0) do
//   begin
//     Result := Result + value(a);
//     Counter := Counter - One;
//   end

void xPascalliVisitor_thirdpass::
visit_xMultiplication(xMultiplication & node)
{
  xCommandBlock* Program = 0;
  xCommandBlock* WhileCommand = 0;
  xVariable*     Result = 0;
  xVariable*     Counter = 0;

  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);

  Side = Right;
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);

  Result = MemManager.New();
  Counter = MemManager.New();
  WhileCommand = new xCommandBlock;

  // Counter:= right;
  Program->addCommand(new xAssignment(Counter, (xExpression*)(Results.pop())));
  Program->addCommand(new xAssignment(Result, new xConstant("0")));
  // Result := 0
  Program->addCommand(new xWhile(new xUnEqual(new xVariable(Counter->Name), new xConstant("0")), WhileCommand));
  WhileCommand->addCommand(new xAssignment
			   (new xVariable(Result->Name), 
			    new xAddition(new xVariable(Result->Name), (xExpression*)(Results.pop()))));
  WhileCommand->addCommand(new xAssignment
			   (new xVariable(Counter->Name), 
			    new xSubtraction(new xVariable(Counter->Name), new xVariable("One"))));

  Results.push(new xVariable(Result->Name));
};




//
// to translate:
// a / b
//
// translation:
//   Result:=0;  
//   Counter:=value(a)-value(b);
//   Correction := b
//   while(Counter) # 0) do
//   begin
//     Result := Result + 1;
//     Correction := Correction + b
//     Counter := Counter - b;
//   end
//   Correction := 1 - (Correction - a)
//   Result := Result + Correction
//

void xPascalliVisitor_thirdpass::
visit_xDivision(xDivision & node)
{
  xCommandBlock* Program = 0;
  xCommandBlock* DivisionLoop = 0;
  xVariable*     Correction = 0;
  xVariable*     Result = 0;
  xVariable*     Counter = 0;
  xVariable*     Dividend = 0;
  xVariable*     Divisor = 0;

  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);

  Side = Right;
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  Dividend = MemManager.New();
  Divisor = MemManager.New();
  Program->addCommand(new xAssignment(Dividend, (xExpression*)(Results.pop())));
  Program->addCommand(new xAssignment(Divisor, (xExpression*)(Results.pop())));

  Result = MemManager.New();
  Counter = MemManager.New();
  Correction = MemManager.New();
  DivisionLoop = new xCommandBlock;

  Program->addCommand(new xAssignment(new xVariable(Correction->Name), new xVariable(Dividend->Name)));
  Program->addCommand(new xAssignment(Counter, new xSubtraction(new xVariable(Divisor->Name), new xVariable(Dividend->Name))));
  Program->addCommand(new xAssignment(Result, new xConstant("0")));
  Program->addCommand(new xWhile(new xUnEqual(new xVariable(Counter->Name), new xConstant("0")), DivisionLoop));
  
  DivisionLoop->addCommand(new xAssignment
			    (new xVariable(Result->Name), 
			     new xAddition(new xVariable(Result->Name), new xVariable("One"))));
  DivisionLoop->addCommand(new xAssignment
			    (new xVariable(Counter->Name), 
			     new xSubtraction(new xVariable(Counter->Name), new xVariable(Dividend->Name))));
  DivisionLoop->addCommand(new xAssignment
			    (new xVariable(Correction->Name), 
			     new xAddition(new xVariable(Correction->Name), new xVariable(Dividend->Name))));

  Program->addCommand(new xAssignment
		       (new xVariable(Correction->Name), 
			new xSubtraction(new xVariable(Correction->Name), new xVariable(Divisor->Name))));
  Program->addCommand(new xAssignment
		       (new xVariable(Correction->Name), 
			new xSubtraction(new xVariable("One"), new xVariable(Correction->Name)))); 
  Program->addCommand(new xAssignment
		       (new xVariable(Result->Name), 
			new xAddition(new xVariable(Result->Name), new xVariable(Correction->Name))));

  Results.push(new xVariable(Result->Name));
};




void xPascalliVisitor_thirdpass::
visit_xConstant(xConstant & node)
{
  xCommandBlock* Program = 0;
  xVariable*     Tmp = 0;
  
  Program = InsertationPoint.pop();
  InsertationPoint.push(Program);

  Tmp = MemManager.New();
  Program->addCommand(new xAssignment(Tmp, &node));

  Results.push(new xVariable(Tmp->Name));
};




void xPascalliVisitor_thirdpass::
visit_xIdentifier(xIdentifier & node)
{
  Results.push(new xVariable(MemManager.getAddress(node)));
  delete(&node);
};



//
// Remarks:
//   We have to distinguish between an array on the left side of an assignment and an array
//   on the right side. If the array is on the left side then we can just assign 
//   aXX[Index->ResultAddress] to node->ResultAddress. If the array is on the right side, we have 
//   to do an assignment temp := aXX[Index->ResultAddress] and assign temp to node->ResultAddress.
//

void xPascalliVisitor_thirdpass::
visit_xArray(xArray & node)
{
  ESide          ThisSide;
  xCommandBlock* Program = 0;
  xVariable*     Tmp = 0;

  ThisSide = Side;
  Side = Right;
  node.Index->AcceptVisitor(*this);

  node.Index = (xExpression*)(Results.pop());
  node.Name = MemManager.getAddress(node);

  switch (ThisSide)
  {
  case Left:
    Results.push(&node);
    break;
  case Right:
    Program = InsertationPoint.pop();
    InsertationPoint.push(Program);
    Tmp = MemManager.New();
    
    Program->addCommand(new xAssignment(Tmp, &node));
    Results.push(new xVariable(Tmp->Name));
    break;
  };
};



