/*
  plic visitor
  Copyright (C) 1999  Achim Kern

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// ---------------------------------------------------------------------------- 
//  Description      : Implementation of the ancestor of all passes forming the
//                     backend of the pascalli to pascallino compiler.
//                     
//                     This visitor proivides function to travers a pascalli-
//                     parsetree and just rebuild the original program.
// ---------------------------------------------------------------------------- 
//  Creation         :
// ---------------------------------------------------------------------------- 
//  Remarks          : 
// ---------------------------------------------------------------------------- 
//  Modifications    : 
//      $Log: pli_visitor.cpp,v $
// Revision 1.2  1999/07/19  14:50:45  thieleek
// first complete revision
//
// Revision 1.1.1.1  1999/07/13  13:35:50  thieleek
// sources Imported
//
// ---------------------------------------------------------------------------- 




#include "plic_Traverser.h"
#include <iostream>



xPascalliTraverser::xPascalliTraverser()
{ 
};




xNode* xPascalliTraverser::
startCompilation(xNode* topnode)
{
  cout << topnode->ResultAddress;
  topnode->AcceptVisitor(*this);
  //  return topnode;
  return Results.pop();
};




void xPascalliTraverser::
visit_xProgramNode(xProgramNode & node)
{
  for (TIndex Counter = 0; Counter < ~(node.Functions); Counter++)
  {
    node.Functions[Counter]->AcceptVisitor(*this);
    node.Functions[Counter] = (xFunction*)(Results.pop());
  };
  Results.push(&node);
};




void xPascalliTraverser::
visit_xFunction(xFunction & node)
{
  for (TIndex Counter = 0; Counter < ~(node.Arguments); Counter++)
  {
    node.Arguments[Counter]->AcceptVisitor(*this);
    node.Arguments[Counter] = (xIdentifier*)(Results.pop());
  };
  node.Command->AcceptVisitor(*this);
  node.Command = (xCommand*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xCommandBlock(xCommandBlock & node)
{
  xCommandBlock* CurrentCommandBlock;

  InsertationPoint.push(new xCommandBlock);
  for (TIndex Counter = 0; Counter < ~(node.CommandList); Counter++)
  {
    node.CommandList[Counter]->AcceptVisitor(*this);
    CurrentCommandBlock = InsertationPoint.pop();
    CurrentCommandBlock->addCommand((xCommand*)(Results.pop()));
    InsertationPoint.push(CurrentCommandBlock);
  };
  Results.push(InsertationPoint.pop());
};




void xPascalliTraverser::
visit_xAssignment(xAssignment & node)
{
  node.Destination->AcceptVisitor(*this);
  node.Source->AcceptVisitor(*this);
  node.Source = (xExpression*)(Results.pop());
  node.Destination = (xIdentifier*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xWhile(xWhile & node)
{
  node.Condition->AcceptVisitor(*this);
  node.Command->AcceptVisitor(*this);
  node.Command = (xCommand*)(Results.pop());
  node.Condition = (xCondition*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xIfStatement(xIfStatement & node)
{
  xCommandBlock* IfCommand = 0;
  xCommandBlock* ElseCommand = 0;

  node.Condition->AcceptVisitor(*this);
  node.Condition = (xCondition*)(Results.pop());

  IfCommand = new xCommandBlock;
  InsertationPoint.push(IfCommand);
  node.IfTwig->AcceptVisitor(*this);
  IfCommand->addCommand((xCommand*)(Results.pop()));
  node.IfTwig = IfCommand;
  InsertationPoint.pop();

  if (node.ElseTwig != 0)
  {
    ElseCommand = new xCommandBlock;
    InsertationPoint.push(ElseCommand);
    node.ElseTwig->AcceptVisitor(*this);
    ElseCommand->addCommand((xCommand*)(Results.pop()));
    node.ElseTwig = ElseCommand;
    InsertationPoint.pop();
  };

  Results.push(&node);
};





void xPascalliTraverser::
visit_xForLoop(xForLoop & node)
{
  node.Counter->AcceptVisitor(*this);
  node.StartValue->AcceptVisitor(*this);
  node.EndValue->AcceptVisitor(*this);
  node.Command->AcceptVisitor(*this);
  node.Command = (xCommand*)(Results.pop());
  node.EndValue = (xExpression*)(Results.pop());
  node.StartValue = (xExpression*)(Results.pop());
  node.Counter = (xIdentifier*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xAddition(xAddition & node)
{
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  node.Right = (xExpression*)(Results.pop());
  node.Left = (xExpression*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xSubtraction(xSubtraction & node)
{
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  node.Right = (xExpression*)(Results.pop());
  node.Left = (xExpression*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xMultiplication(xMultiplication & node)
{
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  node.Right = (xExpression*)(Results.pop());
  node.Left = (xExpression*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xDivision(xDivision & node)
{
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  node.Right = (xExpression*)(Results.pop());
  node.Left = (xExpression*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xConstant(xConstant & node)
{
  Results.push(&node);
};




void xPascalliTraverser::
visit_xIdentifier(xIdentifier & node)
{
  Results.push(&node);
};




void xPascalliTraverser::
visit_xArray(xArray & node)
{
  node.Index->AcceptVisitor(*this);
  node.Index = (xExpression*)(Results.pop());
  
  Results.push(&node);
};




void xPascalliTraverser::
visit_xFunctionCall(xFunctionCall & node)
{
  for (TIndex Counter = 0; Counter < ~(node.Arguments); Counter++)
  {
    node.Arguments[Counter]->AcceptVisitor(*this);
    node.Arguments[Counter] = (xExpression*)(Results.pop());
  };

  Results.push(&node);
};




void xPascalliTraverser::
visit_xAnd(xAnd & node)
{
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  node.Right = (xCondition*)(Results.pop());
  node.Left = (xCondition*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xOr(xOr & node)
{
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  node.Right = (xCondition*)(Results.pop());
  node.Left = (xCondition*)(Results.pop());

  Results.push(&node);
}




void xPascalliTraverser::
visit_xNot(xNot & node)
{
  node.Expression->AcceptVisitor(*this);
  node.Expression = (xCondition*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xGreaterEqual(xGreaterEqual & node)
{
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  node.Right = (xExpression*)(Results.pop());
  node.Left = (xExpression*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xGreater(xGreater & node)
{
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  node.Right = (xExpression*)(Results.pop());
  node.Left = (xExpression*)(Results.pop()); 

  Results.push(&node);
};




void xPascalliTraverser::
visit_xEqual(xEqual & node)
{
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  node.Right = (xExpression*)(Results.pop());
  node.Left = (xExpression*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xSmaller(xSmaller & node)
{
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  node.Right = (xExpression*)(Results.pop());
  node.Left = (xExpression*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xSmallerEqual(xSmallerEqual & node)
{
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  node.Right = (xExpression*)(Results.pop());
  node.Left = (xExpression*)(Results.pop());

  Results.push(&node);
};




void xPascalliTraverser::
visit_xUnEqual(xUnEqual & node)
{
  node.Left->AcceptVisitor(*this);
  node.Right->AcceptVisitor(*this);
  node.Right = (xExpression*)(Results.pop());
  node.Left = (xExpression*)(Results.pop());

  Results.push(&node);
};
