/*
  ramsim main widget
  Copyright (C) 1999  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "mainwidget.hpp"
#include "mainwidget.moc"

#include "bcode.h"

#include <qpushbt.h>
#include <stdio.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qmenubar.h>
#include <qapp.h>
#include <qfiledlg.h>
#include <fstream.h>
#include <qpainter.h>
#include <qmessagebox.h>
#include <qregexp.h>

#include "config.h"
#include "arrow.xpm"
#include "errorcross.xpm"


int cwtextxpos, cwtextypos, cwheight;
QFont *cwfont;


bool IsGreater (const xString str1, const xString str2)
{
  if (~str1 > ~str2) return true;
  else if (~str1 < ~str2) return false;
  else return str1 > str2;
}                                                                               


quickmaker::~quickmaker ()
{
  tv->setAutoUpdate(TRUE);
  tv->update();
};

quickmaker::quickmaker (QListBox *const tv_)
  : tv(tv_)
{
  if (!tv->autoUpdate()) fatal("am already quickmade");
  tv->setAutoUpdate(FALSE);
};


codeitem::codeitem( const char *s, const arrow_mode_type arrow_mode_ )
  : QListBoxItem(), arrow_mode(arrow_mode_)
{
  setText( s );
};

void codeitem::paint( QPainter *p )
{
  QPixmap *mine;
  switch(arrow_mode) {
  default: mine=0; break;
  case GREEN: mine=pixmap_arrow; break;
  case RED: mine=pixmap_errorcross; break;
  };

  if (mine) p->drawPixmap( 5, (height(0)-mine->height())/2, *mine );
  p->setFont(*cwfont);
  p->drawText( cwtextxpos, cwtextypos, text() );
};

int codeitem::height( const QListBox * ) const
{
  return cwheight;
};

int codeitem::width( const QListBox * ) const
{
  return 1; // dunno matter.
};


codew::codew(QWidget *parent,const char *name)
  : QWidget (parent,name)
{
  QVBoxLayout *l = new QVBoxLayout(this);

  QLabel *n = new QLabel ("Code",this);
  n->setAlignment(AlignCenter);
  n->setMinimumSize(n->sizeHint()); n->setMaximumHeight(n->sizeHint().height());
  l->addWidget(n);

  bw = new QListBox (this);
  bw->setMinimumSize(300,300);
  l->addWidget(bw);

  l->activate();
};

void codew::clear()
{
  bw->clear();
};

void codew::changeline (const xRAM &ram, const TIndex line,
			const codeitem::arrow_mode_type arrow_mode)
{
  xByteCode bcode = ram.BCProgram[line];
  QString neu;
  switch (bcode.Code) {
  case BC_STOP:
    neu.sprintf("%5d   STOP",line+1);
    break;
  case BC_MOV_R_R:
    neu.sprintf("%5d   R%s <- R%s",line+1,
		static_cast<const char *const>(bcode.Param1),
		static_cast<const char *const>(bcode.Param2));
    break;
  case BC_MOV_R_RR:
    neu.sprintf("%5d   R%s <- RR%s",line+1,
		static_cast<const char *const>(bcode.Param1),
		static_cast<const char *const>(bcode.Param2));
    break;
  case BC_MOV_RR_R:
    neu.sprintf("%5d   RR%s <- R%s",line+1,
		static_cast<const char *const>(bcode.Param1),
		static_cast<const char *const>(bcode.Param2));
    break;
  case BC_MOV_R_K:
    neu.sprintf("%5d   R%s <- %s",line+1,
		static_cast<const char *const>(bcode.Param1),
		static_cast<const char *const>(bcode.Param2));
    break;
  case BC_ADD:
    neu.sprintf("%5d   R%s <- R%s + R%s",line+1,
		static_cast<const char *const>(bcode.Param1),
		static_cast<const char *const>(bcode.Param2),
		static_cast<const char *const>(bcode.Param3));
    break;
  case BC_SUB:
    neu.sprintf("%5d   R%s <- R%s - R%s",line+1,
		static_cast<const char *const>(bcode.Param1),
		static_cast<const char *const>(bcode.Param2),
		static_cast<const char *const>(bcode.Param3));
    break;
  case BC_GOTO:
    neu.sprintf("%5d   GOTO %s",line+1,
		static_cast<const char *const>(bcode.Param1));
    break;
  case BC_IF_NULL:
    neu.sprintf("%5d   IF R%s = 0 GOTO %s",line+1,
		static_cast<const char *const>(bcode.Param1),
		static_cast<const char *const>(bcode.Param2));
    break;
  case BC_IF_NOT_NULL:
    neu.sprintf("%5d   IF R%s # 0 GOTO %s",line+1,
		static_cast<const char *const>(bcode.Param1),
		static_cast<const char *const>(bcode.Param2));
    break;
  default:
    neu="unknown code";
    break;
  };
  QListBoxItem *elem=new codeitem(neu,arrow_mode);
  if (bw->count()>line) bw->changeItem(elem,line); else bw->insertItem(elem,line);
};

void codew::centerline (const TIndex line)
{
  if (int(line) < bw->topItem()) bw->setTopItem(line);

  // if long program, on STOP command, press STEP, programm will scroll.
  // fix would be use bw->numRowsVisible() but something like that dunno exist
  // need some bigpatch or me is idiot. just ignore this...
  if ( int(line) - bw->numItemsVisible() +2 > bw->topItem() )
    bw->setTopItem( int(line) - bw->numItemsVisible() +2 );
};

int codew::getcurrentline()
{
  return bw->currentItem();
};


regw::regw(QWidget *parent,const char *name)
  : QWidget (parent,name)
{
  QVBoxLayout *l = new QVBoxLayout(this);

  QLabel *n = new QLabel ("Registers",this);
  n->setAlignment(AlignCenter);
  n->setMinimumSize(n->sizeHint()); n->setMaximumHeight(n->sizeHint().height());
  l->addWidget(n);

  b = new QListBox (this);
  b->setFont(QFont("fixed",12));
  b->setMinimumSize(300,300);
  l->addWidget(b);

  l->activate();

  connect(b,SIGNAL(selected(int)),SLOT(selected(int)));
};

void regw::show(xRAM &ram)
{
  int oldtopitem=b->topItem();
  int oldcuritem=b->currentItem();

  b->setAutoUpdate(FALSE);
  b->clear();
  QString neu;

  QRegExp regex("[0-9]+");
  QString dummy;
  unsigned p;
  int len, pos;
  xString neuvalue, aktvalue;

  for (uint x=0;x<~ram.RegisterManager.RegisterList;x++) {
    neuvalue = ram.RegisterManager.RegisterList[x]->GetName();
    for (p=0;p<b->count();p++) {
      dummy=b->text(p);
      pos=regex.match(dummy,0,&len);
      aktvalue=dummy.mid(pos,len);
      if (IsGreater(aktvalue,neuvalue)) break;
    };
    neu.sprintf("R%7s  %18s", static_cast<const char *>(ram.RegisterManager.RegisterList[x]->GetName()),
		static_cast<const char *>(ram.RegisterManager.RegisterList[x]->GetValue()));
    b->insertItem(neu,p);
  };
  if ((oldcuritem>=0) && (oldcuritem < int(b->count()))) b->setCurrentItem(oldcuritem);
  if ((oldtopitem>=0) && (oldtopitem < int(b->count()))) b->setTopItem(oldtopitem);
  b->setAutoUpdate(TRUE);
  b->update();
};

xString regw::selectedreg()
{
  if (b->currentItem()!=-1) {
    QRegExp re("[0-9]+");
    QString parsme(b->text(b->currentItem()));
    int len,pos=re.match(parsme,0,&len);
    return xString(parsme.mid(pos,len));
  }else return xString();
};

void regw::selected ( int index )
{
  QRegExp re("[0-9]+");
  QString parsme(b->text(index));
  int len,pos=re.match(parsme,0,&len);
  emit doubleclicked (xString(parsme.mid(pos,len)));
};


valid::valid ( QWidget * parent, const char * name = 0 )
  : QValidator (parent,name)
{
};

QValidator::State valid::validate ( QString &input, int & ) const
{
  for (unsigned i=0; i<input.length(); i++)
    if (!((input[i]>='0')&&(input[i]<='9'))) return Invalid;

  if ((input.length()>1)&&(input[0]=='0')) return Invalid;

  return Acceptable;
};


regenter::regenter( QWidget *parent, const char *name )
  : QDialog( parent, name, TRUE )
{
  resize(0,0);

  QVBoxLayout *tl = new QVBoxLayout(this,5);

  QGridLayout *v = new QGridLayout(3,2,5);
  tl->addLayout(v,1);

  iv = new valid(this);
  
  QLabel *lab = new QLabel ("Register index",this);
  lab->setMinimumSize(lab->sizeHint());
  v->addWidget(lab,0,0);

  lern = new QLineEdit(this);
  lern->setValidator(iv);
  lern->setMinimumSize(lern->sizeHint());
  lern->setMinimumWidth(400);
  v->addWidget(lern,0,1);

  QLabel *lab2 = new QLabel ("Register value",this);
  lab2->setMinimumSize(lab2->sizeHint());
  v->addWidget(lab2,1,0);

  le = new QLineEdit(this);
  le->setValidator(iv);
  le->setMinimumSize(le->sizeHint());
  v->addWidget(le,1,1);

  QHBoxLayout *hb=new QHBoxLayout(5);
  tl->addLayout(hb,1);

  QPushButton *ok;
  ok = new QPushButton( "Ok", this );
  connect( ok, SIGNAL(clicked()), SLOT(accept()) );
  ok->setMinimumSize(ok->sizeHint());
  hb->addWidget(ok,1);

  QPushButton *cancel;
  cancel = new QPushButton( "Cancel", this );
  connect( cancel, SIGNAL(clicked()), SLOT(reject()) );
  cancel->setMinimumSize(cancel->sizeHint());
  hb->addWidget(cancel,1);

  tl->activate();
}


mainw::~mainw()
{
  delete cwfont;
  delete pixmap_errorcross;
  delete pixmap_arrow;
};

mainw::mainw(QWidget *parent,const char *name)
  : QWidget (parent,name), ram_mode(NOTHING)
{
  resize(0,0);
  
  pixmap_arrow = new QPixmap ((const char **)arrow_xpm);
  pixmap_errorcross = new QPixmap ((const char **)errorcross_xpm);

  cwfont = new QFont("Fixed",12);

  cwheight = pixmap_arrow->height();
  if (pixmap_errorcross->height()>cwheight) cwheight=pixmap_errorcross->height();
  QFontMetrics fm(*cwfont);
  if (fm.height()+4>cwheight) cwheight=fm.height()+4;

  cwtextxpos=pixmap_arrow->width();
  if (pixmap_errorcross->width()>cwtextxpos) cwtextxpos=pixmap_errorcross->width();
  cwtextxpos+=5+5;
  
  cwtextypos=(cwheight+fm.ascent()-1)/2;

  QVBoxLayout *v = new QVBoxLayout(this,5);

  QMenuBar *menubar = new QMenuBar( this );
  QPopupMenu *popup = new QPopupMenu;
  popup->insertItem( "&Open", this, SLOT(file()) );
  popup->insertSeparator();
  popup->insertItem( "&Quit", qApp, SLOT(quit()) );
  menubar->insertItem( "&File", popup );
  QPopupMenu *help = new QPopupMenu;
  help->insertItem( "&About", this, SLOT(about()) );
  help->insertItem( "About &Qt", this, SLOT(aboutQt()) );
  menubar->insertSeparator();
  menubar->insertItem("&Help", help );
  v->setMenuBar( menubar );

  QHBoxLayout *l = new QHBoxLayout(5);
  v->addLayout(l,1);

  cw = new codew(this);
  l->addWidget(cw);

  rw = new regw(this);
  connect(rw,SIGNAL(doubleclicked(xString)),SLOT(doubleclicked(xString)));
  l->addWidget(rw);

  QHBoxLayout *h = new QHBoxLayout(5);
  v->addLayout(h);

  bt1 = new QPushButton ("Interrupt",this);
  bt1->setEnabled(FALSE);
  connect(bt1,SIGNAL(clicked()),SLOT(interrupt()));
  bt1->setMinimumSize(bt1->sizeHint());
  h->addWidget(bt1);

  bt2 = new QPushButton ("Step",this);
  connect(bt2,SIGNAL(clicked()),SLOT(step()));
  bt2->setMinimumSize(bt2->sizeHint());
  h->addWidget(bt2);

  bt3 = new QPushButton ("Until",this);
  connect(bt3,SIGNAL(clicked()),SLOT(until()));
  bt3->setMinimumSize(bt3->sizeHint());
  h->addWidget(bt3);

  bt4 = new QPushButton ("Run",this);
  connect(bt4,SIGNAL(clicked()),SLOT(run()));
  bt4->setMinimumSize(bt4->sizeHint());
  h->addWidget(bt4);

  bt45 = new QPushButton ("Run slowly", this);
  connect(bt45,SIGNAL(clicked()),SLOT(runslowly()));
  bt45->setMinimumSize(bt45->sizeHint());
  h->addWidget(bt45);

  bt5 = new QPushButton ("SetReg",this);
  connect(bt5,SIGNAL(clicked()),SLOT(setreg()));
  bt5->setMinimumSize(bt5->sizeHint());
  h->addWidget(bt5);

  bt6 = new QPushButton ("Reset",this);
  connect(bt6,SIGNAL(clicked()),SLOT(resetall()));
  bt6->setMinimumSize(bt6->sizeHint());
  h->addWidget(bt6);

  h->addStretch(1);

  v->activate();

  if (qApp->argc()>1) openfile(qApp->argv()[1]);
};

void mainw::timerEvent ( QTimerEvent * )
{
  if (runslowlymode) cw->changeline(ram,ram_position,codeitem::NORMAL);

  try{
    ram.Execute(1);
  } catch(xRAMException) {
    ram_mode=ERROR;
  };

  if ((ram_mode==RUNNING) && (!ram.HasTerminated())) {
    ram_position=ram.ReadCommandRegister();

    if (runslowlymode) {
      cw->changeline(ram,ram_position,ram_mode==ERROR ? codeitem::RED : codeitem::GREEN);
      cw->centerline(ram_position);
      rw->show(ram);
    };

    if (ram_position==ram_stopline) interrupt();
  }else{
    interrupt();
  };
};

void mainw::file()
{
  interrupt();
  QString filename (QFileDialog::getOpenFileName(0,0,this));

  if (!filename.isNull()) openfile(filename);
};

void mainw::interrupt ()
{
  killTimers();
  bt1->setEnabled(FALSE);
  bt2->setEnabled(TRUE);
  bt3->setEnabled(TRUE);
  bt4->setEnabled(TRUE);
  bt45->setEnabled(TRUE);
  bt5->setEnabled(TRUE);
  bt6->setEnabled(TRUE);
  if (ram_mode!=NOTHING) {
    cw->changeline(ram,ram_position,ram_mode==ERROR ? codeitem::RED : codeitem::GREEN);
    cw->centerline(ram_position);
  };
  rw->show(ram);
};

void mainw::step ()
{
  if (ram_mode!=NOTHING)
    cw->changeline(ram,ram_position,codeitem::NORMAL);
  //  return;
  switch (ram_mode) {
  case RUNNING: {
    try{
      ram.Execute(1);
    } catch(xRAMException) {
      ram_mode=ERROR;
    };
    if ((ram_mode==RUNNING) && (!ram.HasTerminated())) ram_position=ram.ReadCommandRegister();
    rw->show(ram);
    break;
  };
  case ERROR:
    qApp->beep();
    break;
  case NOTHING:
    qApp->beep();
    break;
  };
  if (ram_mode!=NOTHING) {
    cw->changeline(ram,ram_position,ram_mode==ERROR ? codeitem::RED : codeitem::GREEN);
    cw->centerline(ram_position);
  };
};

void mainw::until ()
{
  if ((ram_mode==RUNNING)&&(cw->getcurrentline()!=-1)) {
    runslowlymode=false;
    startTimer(0);
    bt1->setEnabled(TRUE);
    bt2->setEnabled(FALSE);
    bt3->setEnabled(FALSE);
    bt4->setEnabled(FALSE);
    bt45->setEnabled(FALSE);
    bt5->setEnabled(FALSE);
    bt6->setEnabled(FALSE);
    cw->changeline(ram,ram_position,codeitem::NORMAL);
    ram_stopline=cw->getcurrentline();
  }else{
    qApp->beep();
  };
};

void mainw::run ()
{
  if (ram_mode==RUNNING) {
    runslowlymode=false;
    startTimer(0);
    bt1->setEnabled(TRUE);
    bt2->setEnabled(FALSE);
    bt3->setEnabled(FALSE);
    bt4->setEnabled(FALSE);
    bt45->setEnabled(FALSE);
    bt5->setEnabled(FALSE);
    bt6->setEnabled(FALSE);
    cw->changeline(ram,ram_position,codeitem::NORMAL);
    ram_stopline=99999999; // this limits maximum lines of code in ram to 99999999
  }else{
    qApp->beep();
  };
};

void mainw::runslowly()
{
  if (ram_mode==RUNNING) {
    runslowlymode=true;
    startTimer(500);
    bt1->setEnabled(TRUE);
    bt2->setEnabled(FALSE);
    bt3->setEnabled(FALSE);
    bt4->setEnabled(FALSE);
    bt45->setEnabled(FALSE);
    bt5->setEnabled(FALSE);
    bt6->setEnabled(FALSE);
    ram_stopline=99999999; // this limits maximum lines of code in ram to 99999999
  }else{
    qApp->beep();
  };
};

void mainw::setreg ()
{
  xString x=rw->selectedreg();
  regenter *r = new regenter(this);
  if (x!="") {
    r->lern->setText(QString(x));
    r->le->setText(QString(ram.RegisterManager[x].GetValue()));
  }else{
    r->lern->setText("0");
    r->le->setText("0");
  };
  if (r->exec()) {
    ram.RegisterManager[r->lern->text().latin1()]=r->le->text();
    rw->show(ram);
  };
};

void mainw::resetall()
{
  if (ram_mode==NOTHING) {
    qApp->beep();
  }else{
    ram.Reset();
    codew::quick cqm(cw);
    cw->clear();
    for (TIndex p=0;p<~ram.BCProgram;p++) cw->changeline(ram,p,codeitem::NORMAL);
    if (~ram.BCProgram>0) {
      ram_mode=RUNNING;
      ram_position=0;
      cw->changeline(ram,0,codeitem::GREEN);
    };
    rw->show(ram);
  };
};

void mainw::doubleclicked (xString regnr)
{
  regenter r(this);

  r.lern->setText(QString(regnr));
  r.le->setText(QString(ram.RegisterManager[regnr].GetValue()));
  if (r.exec()) {
    ram.RegisterManager[r.lern->text().latin1()]=r.le->text();
    rw->show(ram);
  };
};

void mainw::about()
{
  QString msg;
  msg+="ramsim version "; msg+=VERSION; msg+=", Copyright (C) 1999  Erik Thiele\n";
  msg+="ramsim comes with ABSOLUTELY NO WARRANTY; for details read file LICENSE\n"
    "This is free software, and you are welcome to redistribute it\n"
    "under certain conditions; for details read file LICENSE\n"
    "\n"
    "ramsim is simulating RAM programs\n"
    "it is a part of the theosim package";
  QMessageBox::about( this, "ramsim", msg );
};

void mainw::aboutQt()
{
  QMessageBox::aboutQt( this, "ramsim" );
};

void mainw::openfile(QString name)
{
  ifstream stream(name);
  stream >> ramprog;
  ram.Load(ramprog);
  codew::quick cqm(cw);
  cw->clear();
  for (TIndex p=0;p<~ram.BCProgram;p++) cw->changeline(ram,p,codeitem::NORMAL);
  if (~ram.BCProgram>0) {
    ram_mode=RUNNING;
    ram_position=0;
    cw->changeline(ram,0,codeitem::GREEN);
  };
  rw->show(ram);
};

QPixmap *pixmap_arrow, *pixmap_errorcross;
