/*
  ramsim main widget
  Copyright (C) 1999  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __MAINWIDGET_HPP
#define __MAINWIDGET_HPP

#include <ram.h>

#include <qwidget.h>
#include <qlistbox.h>
#include <qdialog.h>
#include <qlineedit.h>
#include <qvalidator.h>


class quickmaker {
public:
  quickmaker (const quickmaker &);
  quickmaker & operator= (const quickmaker &);
  ~quickmaker ();
  quickmaker (QListBox *const tv_);
private:
  QListBox *tv;
};

class codeitem : public QListBoxItem {
public:
  enum arrow_mode_type { NORMAL, GREEN, RED };

  codeitem( const char *s, const arrow_mode_type arrow_mode_ );
  
protected:
  void paint( QPainter *p );
  int height( const QListBox * ) const;
  int width( const QListBox * ) const;

private:
  arrow_mode_type arrow_mode;
};

class codew : public QWidget {
  Q_OBJECT
public:
  struct quick : public quickmaker {
    quick (codew *const cw) : quickmaker (cw->bw) {};
  };
  friend class codew::quick;

  codew(QWidget *parent=0,const char *name=0);

  void clear();
  void changeline (const xRAM &ram, const TIndex line, const codeitem::arrow_mode_type arrow_mode);
  void centerline (const TIndex line);
  int getcurrentline(); // -1 if nothing selected

private:
  QListBox *bw;
};

class regw : public QWidget {
  Q_OBJECT
public:
  regw(QWidget *parent=0,const char *name=0);
  
  void show(xRAM &ram);
  xString selectedreg(); // returns "" if none selected

private slots:
  void selected ( int index );
signals:
  void doubleclicked (xString regnr);
  void valuechanged (xNamedRegister neu);
  
private:
  QListBox *b;
};

class valid : public QValidator {
  Q_OBJECT
public:
  valid ( QWidget * parent, const char * name = 0 );
  State validate ( QString &input, int &pos ) const;
};

class regenter : public QDialog {
  Q_OBJECT
public:
  regenter( QWidget *parent=0, const char *name=0 );

  valid *iv;
  QLineEdit *le, *lern;
};

class mainw : public QWidget {
  Q_OBJECT
public:
  ~mainw ();
  mainw (QWidget *parent=0,const char *name=0);

protected:
  void timerEvent ( QTimerEvent * );

private slots:
  void file ();
  void interrupt ();
  void step ();
  void until ();
  void run ();
  void runslowly ();
  void setreg ();
  void resetall ();
  void doubleclicked (xString regnr);
  void about();
  void aboutQt();

private:
  void openfile(QString name);

  QPushButton *bt1, *bt2, *bt3, *bt4, *bt45, *bt5, *bt6;
  codew *cw;
  regw *rw;
  xRAM ram;
  xRAMProgram ramprog;
  enum ram_mode_type { RUNNING, ERROR, NOTHING };
  ram_mode_type ram_mode;
  TIndex ram_position, ram_stopline;
  bool runslowlymode;
};

extern QPixmap *pixmap_arrow, *pixmap_errorcross;

#endif
