/*
  theosim main widget
  Copyright (C) 1999  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __MW_HPP
#define __MW_HPP

#include <qwidget.h>
#include <qmenubar.h>
#include <qwidgetstack.h>
#include <qlist.h>
#include <qmultilinedit.h>
#include <qlabel.h>
#include <qpixmap.h>
#include <qdialog.h>
#include <qsocketnotifier.h>
#include <sys/types.h>
#include <sys/wait.h>

class external : public QDialog {
  Q_OBJECT
public:
  ~external ();
  external (const char *messagestring,
	    const char *filename, char *const *argvlist, QWidget *parent=0,const char *name=0);

  enum result_t { interrupted, success, failure };
  result_t result;

private:
  QLabel *message, *status;
  QMultiLineEdit *exbox;
  int output[2], error[2];
  QSocketNotifier *outputnotify, *errornotify;
  QPushButton *ok, *interrupt;

  void terminatechild();
  int childtermstatus;
  pid_t childpid;

private slots:
  void activated(int hand);
  void okbut();
  void interruptbut();
};

class bgwdg : public QWidget {
  Q_OBJECT
public:
  bgwdg (QWidget *parent=0,const char *name=0);
  void paintEvent ( QPaintEvent * );
};

enum filetypes { PASCALLI, PASCALLINO, RAM };

class buffer : public QWidget {
  Q_OBJECT
public:
  buffer (QWidget *parent=0, const char *name=0);

  bool changed;
  QMultiLineEdit *editor;
  QLabel *status;
  QString filename;
  filetypes filetype;

  void updatestatus();

private slots:
  void textChanged();
};

class mainw : public QWidget {
  Q_OBJECT
public:
  ~mainw ();
  mainw (QWidget *parent=0,const char *name=0);
private slots:
  void buffersel (int id);
  void fileopen();
  void filesave();
  void fileclose();
  void filequit();
  void pascalli2pascallino();
  void pascallino2ram();
  void ramsim();
  void about();
  void aboutQt();
protected:
  void closeEvent( QCloseEvent *e );
private:
  buffer *findbuffer (const QString filename);
  void realopen(QString filename);

  QMenuBar *menubar;
  QPopupMenu *bufferpop, *tasks, *filepop;
  QWidgetStack *editstack;
  int id_file_close, id_buffer, id_tasks, id_file_save;
  int id_plic,id_ploc,id_ramsim;
};

extern QPixmap *pixmap_theosim;

#endif
